<html>
    <head>
        <script src="https://cdn.plot.ly/plotly-latest.min.js"></script>
        <link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.1/css/bootstrap.min.css">
    </head>
    <h1>
        GeneWalk summary
    </h1>
    <h2>
        Global analyses
    </h2>
    <div class="row">
        {{ GLOBAL_RESULTS }}}
    </div>
    <h2>
        Gene-specific results
    </h2>
    {{ GENE_RESULTS }}
</html>
